@@ -18,6 +18,7 @@ import cherrypy
 import hashlib
 import math
 import os
+import pathlib
 import re
 import urllib
 
@@ -25,10 +26,11 @@ from girder.api import access, filter_logging
 from girder.api.v1.item import Item as ItemResource
 from girder.api.describe import autoDescribeRoute, describeRoute, Description
 from girder.api.rest import filtermodel, loadmodel, setRawResponse, setResponseHeader
+from girder.constants import AccessType
 from girder.exceptions import RestException
-from girder.models.model_base import AccessType
 from girder.models.file import File
 from girder.models.item import Item
+from girder.utility.progress import setResponseTimeLimit
 
 from large_image.constants import TileInputUnits
 from large_image.exceptions import TileGeneralException
@@ -86,8 +88,8 @@ def _handleETag(key, item, *args, **kwargs):
     conditions = [str(x) for x in cherrypy.request.headers.elements('If-None-Match') or []]
     if conditions == ['*'] or etag in conditions:
         raise cherrypy.HTTPRedirect([], 304)
-    # Explicitly set a max-ago to recheck the cahe after a while
-    setResponseHeader('Cache-control', 'max-age=600')
+    # Explicitly set a max-age to recheck the cahe after a while
+    setResponseHeader('Cache-control', 'max-age=3600')
 
 
 class TilesItemResource(ItemResource):
@@ -99,6 +101,7 @@ class TilesItemResource(ItemResource):
 
         self.resourceName = 'item'
         apiRoot.item.route('POST', (':itemId', 'tiles'), self.createTiles)
+        apiRoot.item.route('POST', (':itemId', 'tiles', 'convert'), self.convertImage)
         apiRoot.item.route('GET', (':itemId', 'tiles'), self.getTilesInfo)
         apiRoot.item.route('DELETE', (':itemId', 'tiles'), self.deleteTiles)
         apiRoot.item.route('GET', (':itemId', 'tiles', 'thumbnail'),
@@ -139,10 +142,9 @@ class TilesItemResource(ItemResource):
 
     @describeRoute(
         Description('Create a large image for this item.')
-        .param('itemId', 'The ID of the item.', paramType='path')
-        .param('fileId', 'The ID of the source file containing the image. '
-                         'Required if there is more than one file in the item.',
-               required=False)
+        .param('itemId', 'The source item.', paramType='path')
+        .param('fileId', 'The source file containing the image.  Required if '
+               'there is more than one file in the item.', required=False)
         .param('force', 'Always use a job to create the large image.',
                dataType='boolean', default=False, required=False)
         .param('notify', 'If a job is required to create the large image, '
@@ -164,11 +166,17 @@ class TilesItemResource(ItemResource):
                'quality', dataType='int', required=False)
         .param('cr', 'JP2K target compression ratio where 1 is lossless',
                dataType='int', required=False)
+        .param('concurrent', 'Suggested number of maximum concurrent '
+               'processes to use during conversion.  Values <= 0 use the '
+               'number of logical cpus less that value.  Default is -2.',
+               dataType='int', required=False)
     )
     @access.user
     @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.WRITE)
     @filtermodel(model='job', plugin='jobs')
     def createTiles(self, item, params):
+        if 'concurrent' in params:
+            params['_concurrency'] = params.pop('concurrent')
         largeImageFileId = params.get('fileId')
         if largeImageFileId is None:
             files = list(Item().childFiles(item=item, limit=2))
@@ -190,6 +198,69 @@ class TilesItemResource(ItemResource):
         except TileGeneralException as e:
             raise RestException(e.args[0])
 
+    @describeRoute(
+        Description('Create a new large image item based on an existing item')
+        .notes('This can be used to make an item that is a different internal '
+               'format than the original item.')
+        .param('itemId', 'The source item.', paramType='path')
+        .param('fileId', 'The source file containing the image.  Required if '
+               'there is more than one file in the item.', required=False)
+        .param('folderId', 'The destination folder.', required=False)
+        .param('name', 'A new name for the output item.', required=False)
+        .param('localJob', 'If true, run as a local job; if false, run via '
+               'the remote worker', dataType='boolean', default=True,
+               required=False)
+        .param('tileSize', 'Tile size', dataType='int', default=256,
+               required=False)
+        .param('onlyFrame', 'Only convert a specific 0-based frame of a '
+               'multiframe file.  If not specified, all frames are converted.',
+               dataType='int', required=False)
+        .param('format', 'File format', required=False,
+               enum=['tiff', 'aperio'])
+        .param('compression', 'Internal compression format', required=False,
+               enum=['none', 'jpeg', 'deflate', 'lzw', 'zstd', 'packbits', 'webp', 'jp2k'])
+        .param('quality', 'JPEG compression quality where 0 is small and 100 '
+               'is highest quality', dataType='int', default=90,
+               required=False)
+        .param('level', 'Compression level for deflate (zip) or zstd.',
+               dataType='int', required=False)
+        .param('predictor', 'Predictor for deflate (zip) or lzw.',
+               required=False, enum=['none', 'horizontal', 'float', 'yes'])
+        .param('psnr', 'JP2K compression target peak-signal-to-noise-ratio '
+               'where 0 is lossless and otherwise higher numbers are higher '
+               'quality', dataType='int', required=False)
+        .param('cr', 'JP2K target compression ratio where 1 is lossless',
+               dataType='int', required=False)
+        .param('concurrent', 'Suggested number of maximum concurrent '
+               'processes to use during conversion.  Values <= 0 use the '
+               'number of logical cpus less that value.  Default is -2.',
+               dataType='int', required=False)
+    )
+    @access.user
+    @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.READ)
+    @filtermodel(model='job', plugin='jobs')
+    def convertImage(self, item, params):
+        if 'concurrent' in params:
+            params['_concurrency'] = params.pop('concurrent')
+        largeImageFileId = params.get('fileId')
+        if largeImageFileId is None:
+            files = list(Item().childFiles(item=item, limit=2))
+            if len(files) == 1:
+                largeImageFileId = str(files[0]['_id'])
+        if not largeImageFileId:
+            raise RestException('Missing "fileId" parameter.')
+        largeImageFile = File().load(largeImageFileId, force=True, exc=True)
+        user = self.getCurrentUser()
+        token = self.getCurrentToken()
+        params.pop('notify', None)
+        localJob = self.boolParam('localJob', params, default=True)
+        params.pop('localJob', None)
+        try:
+            return self.imageItemModel.convertImage(
+                item, largeImageFile, user, token, localJob=localJob, **params)
+        except TileGeneralException as e:
+            raise RestException(e.args[0])
+
     @classmethod
     def _parseTestParams(cls, params):
         _adjustParams(params)
@@ -261,7 +332,7 @@ class TilesItemResource(ItemResource):
         except TileGeneralException as e:
             raise RestException(e.args[0], code=400)
 
-    def _setContentDisposition(self, item, contentDisposition, mime, subname):
+    def _setContentDisposition(self, item, contentDisposition, mime, subname, fullFilename=None):
         """
         If requested, set the content disposition and a suggested file name.
 
@@ -271,15 +342,20 @@ class TilesItemResource(ItemResource):
         :param mime: the mimetype of the output image.  Used for the filename
             suffix.
         :param subname: a subname to append to the item name.
+        :param fullFilename: if specified, use this instead of the item name
+            and the subname.
         """
         if (not item or not item.get('name') or
                 mime not in MimeTypeExtensions or
                 contentDisposition not in ('inline', 'attachment')):
             return
-        filename = os.path.splitext(item['name'])[0]
-        if subname:
-            filename += '-' + subname
-        filename += '.' + MimeTypeExtensions[mime]
+        if fullFilename:
+            filename = fullFilename
+        else:
+            filename = os.path.splitext(item['name'])[0]
+            if subname:
+                filename += '-' + subname
+            filename += '.' + MimeTypeExtensions[mime]
         if not isinstance(filename, str):
             filename = filename.decode('utf8', 'ignore')
         safeFilename = filename.encode('ascii', 'ignore').replace(b'"', b'')
@@ -558,7 +634,6 @@ class TilesItemResource(ItemResource):
     @loadmodel(model='item', map={'itemId': 'item'}, level=AccessType.WRITE)
     def deleteTiles(self, item, params):
         deleted = self.imageItemModel.delete(item)
-        # TODO: a better response
         return {
             'deleted': deleted
         }
@@ -585,11 +660,16 @@ class TilesItemResource(ItemResource):
         .param('frame', 'For multiframe images, the 0-based frame number.  '
                'This is ignored on non-multiframe images.', required=False,
                dataType='int')
-        .param('encoding', 'Thumbnail output encoding', required=False,
-               enum=['JPEG', 'PNG', 'TIFF'], default='JPEG')
+        .param('encoding', 'Output image encoding.  TILED generates a tiled '
+               'tiff without the upper limit on image size the other options '
+               'have.  For geospatial sources, TILED will also have '
+               'appropriate tagging.', required=False,
+               enum=['JPEG', 'PNG', 'TIFF', 'TILED'], default='JPEG')
         .param('contentDisposition', 'Specify the Content-Disposition response '
                'header disposition-type value.', required=False,
                enum=['inline', 'attachment'])
+        .param('contentDispositionFilename', 'Specify the filename used in '
+               'the Content-Disposition response header.', required=False)
         .produces(ImageMimeTypes)
         .errorResponse('ID was invalid.')
         .errorResponse('Read access was denied for the item.', 403)
@@ -609,6 +689,7 @@ class TilesItemResource(ItemResource):
             ('encoding', str),
             ('style', str),
             ('contentDisposition', str),
+            ('contentDispositionFileName', str)
         ])
         _handleETag('getTilesThumbnail', item, params)
         try:
@@ -621,7 +702,8 @@ class TilesItemResource(ItemResource):
             return result
         thumbData, thumbMime = result
         self._setContentDisposition(
-            item, params.get('contentDisposition'), thumbMime, 'thumbnail')
+            item, params.get('contentDisposition'), thumbMime, 'thumbnail',
+            params.get('contentDispositionFilename'))
         setResponseHeader('Content-Type', thumbMime)
         setRawResponse()
         return thumbData
@@ -683,8 +765,11 @@ class TilesItemResource(ItemResource):
         .param('frame', 'For multiframe images, the 0-based frame number.  '
                'This is ignored on non-multiframe images.', required=False,
                dataType='int')
-        .param('encoding', 'Output image encoding', required=False,
-               enum=['JPEG', 'PNG', 'TIFF'], default='JPEG')
+        .param('encoding', 'Output image encoding.  TILED generates a tiled '
+               'tiff without the upper limit on image size the other options '
+               'have.  For geospatial sources, TILED will also have '
+               'appropriate tagging.', required=False,
+               enum=['JPEG', 'PNG', 'TIFF', 'TILED'], default='JPEG')
         .param('jpegQuality', 'Quality used for generating JPEG images',
                required=False, dataType='int', default=95)
         .param('jpegSubsampling', 'Chroma subsampling used for generating '
@@ -697,14 +782,16 @@ class TilesItemResource(ItemResource):
                      'tiff_adobe_deflate'])
         .param('style', 'JSON-encoded style string', required=False)
         .param('resample', 'If false, an existing level of the image is used '
-               'for the histogram.  If true, the internal values are '
+               'for the region.  If true, the internal values are '
                'interpolated to match the specified size as needed.  0-3 for '
                'a specific interpolation method (0-nearest, 1-lanczos, '
                '2-bilinear, 3-bicubic)', required=False,
-               enum=['false', 'true', '0', '1', '2', '3'], default='false')
+               enum=['false', 'true', '0', '1', '2', '3'])
         .param('contentDisposition', 'Specify the Content-Disposition response '
                'header disposition-type value.', required=False,
                enum=['inline', 'attachment'])
+        .param('contentDispositionFilename', 'Specify the filename used in '
+               'the Content-Disposition response header.', required=False)
         .produces(ImageMimeTypes)
         .errorResponse('ID was invalid.')
         .errorResponse('Read access was denied for the item.', 403)
@@ -738,8 +825,10 @@ class TilesItemResource(ItemResource):
             ('style', str),
             ('resample', 'boolOrInt'),
             ('contentDisposition', str),
+            ('contentDispositionFileName', str)
         ])
         _handleETag('getTilesRegion', item, params)
+        setResponseTimeLimit(86400)
         try:
             regionData, regionMime = self.imageItemModel.getRegion(
                 item, **params)
@@ -748,8 +837,23 @@ class TilesItemResource(ItemResource):
         except ValueError as e:
             raise RestException('Value Error: %s' % e.args[0])
         self._setContentDisposition(
-            item, params.get('contentDisposition'), regionMime, 'region')
+            item, params.get('contentDisposition'), regionMime, 'region',
+            params.get('contentDispositionFilename'))
         setResponseHeader('Content-Type', regionMime)
+        if isinstance(regionData, pathlib.Path):
+            BUF_SIZE = 65536
+
+            def stream():
+                try:
+                    with regionData.open('rb') as f:
+                        while True:
+                            data = f.read(BUF_SIZE)
+                            if not data:
+                                break
+                            yield data
+                finally:
+                    regionData.unlink()
+            return stream
         setRawResponse()
         return regionData
 
@@ -903,6 +1007,8 @@ class TilesItemResource(ItemResource):
         .param('contentDisposition', 'Specify the Content-Disposition response '
                'header disposition-type value.', required=False,
                enum=['inline', 'attachment'])
+        .param('contentDispositionFilename', 'Specify the filename used in '
+               'the Content-Disposition response header.', required=False)
         .produces(ImageMimeTypes)
         .errorResponse('ID was invalid.')
         .errorResponse('Read access was denied for the item.', 403)
@@ -922,6 +1028,7 @@ class TilesItemResource(ItemResource):
             ('encoding', str),
             ('style', str),
             ('contentDisposition', str),
+            ('contentDispositionFileName', str)
         ])
         _handleETag('getAssociatedImage', item, image, params)
         try:
@@ -932,7 +1039,8 @@ class TilesItemResource(ItemResource):
             return result
         imageData, imageMime = result
         self._setContentDisposition(
-            item, params.get('contentDisposition'), imageMime, image)
+            item, params.get('contentDisposition'), imageMime, image,
+            params.get('contentDispositionFilename'))
         setResponseHeader('Content-Type', imageMime)
         setRawResponse()
         return imageData
