@@ -14,14 +14,14 @@ from girder.models.setting import Setting
 from girder.models.token import Token
 from girder.models.user import User
 
+from girder_jobs.constants import JobStatus
 from girder_jobs.models.job import Job
 
-from girder_large_image.models.image_item import ImageItem
-
 from large_image import getTileSource
 from girder_large_image import constants
 from girder_large_image import getGirderTileSource
 from girder_large_image import loadmodelcache
+from girder_large_image.models.image_item import ImageItem
 
 from . import girder_utilities as utilities
 
@@ -134,7 +134,7 @@ def _createTestTiles(server, admin, params=None, info=None, error=None):
     return infoDict
 
 
-def _postTileViaHttp(server, admin, itemId, fileId, jobAction=None, data=None):
+def _postTileViaHttp(server, admin, itemId, fileId, jobAction=None, data=None, convert=False):
     """
     When we know we need to process a job, we have to use an actual http
     request rather than the normal simulated request to cherrypy.  This is
@@ -208,7 +208,7 @@ def _postTileViaHttp(server, admin, itemId, fileId, jobAction=None, data=None):
 @pytest.mark.plugin('large_image')
 def testTilesFromPTIF(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     fileId = str(file['_id'])
     # We should already have tile information.  Ask to delete it so we can
@@ -327,7 +327,7 @@ def testTilesFromPTIF(server, admin, fsAssetstore):
 def testTilesFromTest(server, admin, fsAssetstore):
     publicFolder = utilities.namedFolder(admin, 'Public')
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     items = [{'itemId': str(file['itemId']), 'fileId': str(file['_id'])}]
     # We should already have tile information.  Ask to delete it so we can
     # do other tests
@@ -528,7 +528,7 @@ def testTilesWithUnicodeName(server, admin, fsAssetstore):
     # Unicode file names shouldn't cause problems when accessing ptifs.
     # This requires an appropriate version of the python libtiff module.
     name = '\u0441\u043b\u0430\u0439\u0434.ptif'
-    origpath = utilities.externaldata('data/sample_image.ptif.sha512')
+    origpath = utilities.datastore.fetch('sample_image.ptif')
     altpath = os.path.join(os.path.dirname(origpath), name)
     if os.path.exists(altpath):
         os.unlink(altpath)
@@ -568,7 +568,7 @@ def testTilesFromBadFiles(boundServer, admin, fsAssetstore, girderWorker):
 @pytest.mark.plugin('large_image')
 def testThumbnails(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     fileId = str(file['_id'])
     # We should already have tile information.  Ask to delete it so we can
@@ -667,7 +667,7 @@ def testThumbnails(server, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testContentDisposition(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
 
     params = {'encoding': 'PNG', 'width': 200}
@@ -700,11 +700,28 @@ def testContentDisposition(server, admin, fsAssetstore):
         'largeImageThumbnail' in resp.headers['Content-Disposition'])
 
 
+@pytest.mark.usefixtures('unbindLargeImage')
+@pytest.mark.plugin('large_image')
+def testContentDispositionFilename(server, admin, fsAssetstore):
+    file = utilities.uploadExternalFile(
+        'sample_image.ptif', admin, fsAssetstore)
+    itemId = str(file['itemId'])
+
+    params = {'encoding': 'PNG', 'width': 200}
+    path = '/item/%s/tiles/thumbnail' % itemId
+    params['contentDisposition'] = 'attachment'
+    params['contentDispositionFilename'] = 'sample.jpg'
+    resp = server.request(path=path, user=admin, isJson=False, params=params)
+    assert utilities.respStatus(resp) == 200
+    assert resp.headers['Content-Disposition'].startswith('attachment')
+    assert resp.headers['Content-Disposition'].endswith('.jpg')
+
+
 @pytest.mark.usefixtures('unbindLargeImage')
 @pytest.mark.plugin('large_image')
 def testRegions(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     # Get metadata to use in our tests
     resp = server.request(path='/item/%s/tiles' % itemId, user=admin)
@@ -783,12 +800,21 @@ def testRegions(server, admin, fsAssetstore):
     assert width == 500
     assert height == 375
 
+    # Get a tiled image
+    params = {'regionWidth': 1000, 'regionHeight': 1000,
+              'left': 48000, 'top': 3000, 'encoding': 'TILED'}
+    resp = server.request(path='/item/%s/tiles/region' % itemId,
+                          user=admin, isJson=False, params=params)
+    assert utilities.respStatus(resp) == 200
+    image = origImage = utilities.getBody(resp, text=False)
+    assert image[:len(utilities.BigTIFFHeader)] == utilities.BigTIFFHeader
+
 
 @pytest.mark.usefixtures('unbindLargeImage')
 @pytest.mark.plugin('large_image')
 def testPixel(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
 
     # Test bad parameters
@@ -823,7 +849,7 @@ def testPixel(server, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testGetTileSource(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     # We should have access via getGirderTileSource
     source = getGirderTileSource(itemId, user=admin)
@@ -831,19 +857,11 @@ def testGetTileSource(server, admin, fsAssetstore):
     assert image[:len(utilities.PNGHeader)] == utilities.PNGHeader
 
     # We can also use a file with getTileSource.  The user is ignored.
-    imagePath = utilities.externaldata('data/sample_image.ptif.sha512')
+    imagePath = utilities.datastore.fetch('sample_image.ptif')
     source = getTileSource(imagePath, user=admin, encoding='PNG')
     image, mime = source.getThumbnail(encoding='JPEG', width=200)
     assert image[:len(utilities.JPEGHeader)] == utilities.JPEGHeader
 
-    # Test the level0 thumbnail code path
-    image, mime = source.getThumbnail(
-        encoding='PNG', width=200, height=100, levelZero=True, fill='blue')
-    assert image[:len(utilities.PNGHeader)] == utilities.PNGHeader
-    (width, height) = struct.unpack('!LL', image[16:24])
-    assert width == 200
-    assert height == 100
-
 
 @pytest.mark.usefixtures('unbindLargeImage')
 @pytest.mark.plugin('large_image')
@@ -851,7 +869,7 @@ def testTilesLoadModelCache(server, admin, fsAssetstore):
     loadmodelcache.invalidateLoadModelCache()
     token = str(Token().createToken(admin)['_id'])
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     # Now the tile request should tell us about the file.  These are
     # specific to our test file
@@ -868,7 +886,7 @@ def testTilesLoadModelCache(server, admin, fsAssetstore):
 def testTilesModelLookupCache(server, user, admin, fsAssetstore):
     User().load = mock.Mock(wraps=User().load)
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     token = str(Token().createToken(user)['_id'])
     lastCount = User().load.call_count
@@ -887,7 +905,7 @@ def testTilesModelLookupCache(server, user, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testTilesDZIEndpoints(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(path='/item/%s/tiles' % itemId, user=admin)
     assert utilities.respStatus(resp) == 200
@@ -1013,7 +1031,7 @@ def testTilesAfterCopyItem(boundServer, admin, fsAssetstore, girderWorker):
 @pytest.mark.plugin('large_image')
 def testTilesAutoSetOption(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore,
+        'sample_image.ptif', admin, fsAssetstore,
         name='sample_image.PTIF')
     itemId = str(file['itemId'])
     # We should already have tile information.
@@ -1022,7 +1040,7 @@ def testTilesAutoSetOption(server, admin, fsAssetstore):
     # Turn off auto-set and try again
     Setting().set(constants.PluginSettings.LARGE_IMAGE_AUTO_SET, 'false')
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(path='/item/%s/tiles' % itemId, user=admin)
     assert utilities.respStatus(resp) == 400
@@ -1030,7 +1048,7 @@ def testTilesAutoSetOption(server, admin, fsAssetstore):
     # Turn it back on
     Setting().set(constants.PluginSettings.LARGE_IMAGE_AUTO_SET, 'true')
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(path='/item/%s/tiles' % itemId, user=admin)
     assert utilities.respStatus(resp) == 200
@@ -1040,7 +1058,7 @@ def testTilesAutoSetOption(server, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testTilesAssociatedImages(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
 
     resp = server.request(path='/item/%s/tiles/images' % itemId, user=admin)
@@ -1079,7 +1097,7 @@ def testTilesAssociatedImages(server, admin, fsAssetstore):
 
     # Test with an image that doesn't have associated images
     file = utilities.uploadExternalFile(
-        'data/sample_Easy1.png.sha512', admin, fsAssetstore)
+        'sample_Easy1.png', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(path='/item/%s/tiles' % itemId, method='POST', user=admin)
     assert utilities.respStatus(resp) == 200
@@ -1098,7 +1116,7 @@ def testTilesAssociatedImages(server, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testTilesWithFrameNumbers(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample.ome.tif.sha512', admin, fsAssetstore)
+        'sample.ome.tif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     # Test that we can get frames via either tiles/zxy or tiles/fzxy and
     # that the frames are different
@@ -1129,7 +1147,7 @@ def testTilesWithFrameNumbers(server, admin, fsAssetstore):
 @pytest.mark.plugin('large_image')
 def testTilesHistogram(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(
         path='/item/%s/tiles/histogram' % itemId,
@@ -1138,13 +1156,34 @@ def testTilesHistogram(server, admin, fsAssetstore):
     assert len(resp.json[0]['hist']) == 256
     assert resp.json[1]['samples'] == 2801664
     assert resp.json[1]['hist'][128] == 176
+    # A second query will fetch it from cache
+    resp = server.request(
+        path='/item/%s/tiles/histogram' % itemId,
+        params={'width': 2048, 'height': 2048, 'resample': False})
+    assert len(resp.json) == 3
+    assert len(resp.json[0]['hist']) == 256
+
+
+@pytest.mark.usefixtures('unbindLargeImage')
+@pytest.mark.plugin('large_image')
+def testTilesHistogramWithRange(server, admin, fsAssetstore):
+    file = utilities.uploadExternalFile(
+        'sample_image.ptif', admin, fsAssetstore)
+    itemId = str(file['itemId'])
+    resp = server.request(
+        path='/item/%s/tiles/histogram' % itemId,
+        params={'width': 2048, 'height': 2048, 'resample': False, 'rangeMin': 10, 'rangeMax': 240})
+    assert len(resp.json) == 3
+    assert len(resp.json[0]['hist']) == 256
+    assert resp.json[1]['samples'] == 685979
+    assert resp.json[1]['hist'][128] == 186
 
 
 @pytest.mark.usefixtures('unbindLargeImage')
 @pytest.mark.plugin('large_image')
 def testTilesInternalMetadata(server, admin, fsAssetstore):
     file = utilities.uploadExternalFile(
-        'data/sample_image.ptif.sha512', admin, fsAssetstore)
+        'sample_image.ptif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     resp = server.request(path='/item/%s/tiles/internal_metadata' % itemId)
     assert resp.json['tilesource'] == 'tiff'
@@ -1173,7 +1212,7 @@ def testTilesFromMultipleDotName(boundServer, admin, fsAssetstore, girderWorker)
 @pytest.mark.plugin('large_image')
 def testTilesForcedConversion(boundServer, admin, fsAssetstore, girderWorker):
     file = utilities.uploadExternalFile(
-        'data/landcover_sample_1000.tif.sha512', admin, fsAssetstore)
+        'landcover_sample_1000.tif', admin, fsAssetstore)
     itemId = str(file['itemId'])
     fileId = str(file['_id'])
     # We should already have tile information.  Ask to delete it so we can
@@ -1198,3 +1237,66 @@ def testTilesFromWithOptions(boundServer, admin, fsAssetstore, girderWorker):
     assert tileMetadata['sizeX'] == 10000
     assert tileMetadata['sizeY'] == 5000
     assert tileMetadata['levels'] == 5
+
+
+@pytest.mark.usefixtures('unbindLargeImage')
+@pytest.mark.plugin('large_image')
+def testTilesConvertLocal(boundServer, admin, fsAssetstore):
+    file = utilities.uploadTestFile('grey10kx5k.tif', admin, fsAssetstore)
+    itemId = str(file['itemId'])
+
+    headers = {
+        'Accept': 'application/json',
+        'Girder-Token': str(Token().createToken(admin)['_id'])
+    }
+    req = requests.post('http://127.0.0.1:%d/api/v1/item/%s/tiles/convert' % (
+        boundServer.boundPort, itemId), headers=headers)
+    assert req.status_code == 200
+    job = req.json()
+    while job['status'] not in (JobStatus.SUCCESS, JobStatus.ERROR, JobStatus.CANCELED):
+        time.sleep(0.1)
+        job = Job().load(job['_id'], force=True)
+    item = Item().findOne({'name': 'grey10kx5k.tiff'}, sort=[('created', SortDir.DESCENDING)])
+    itemId = item['_id']
+    tileMetadata = ImageItem().getMetadata(item)
+    assert tileMetadata['tileWidth'] == 256
+    assert tileMetadata['tileHeight'] == 256
+    assert tileMetadata['sizeX'] == 10000
+    assert tileMetadata['sizeY'] == 5000
+    assert tileMetadata['levels'] == 7
+    assert tileMetadata['magnification'] is None
+    assert tileMetadata['mm_x'] is None
+    assert tileMetadata['mm_y'] is None
+    _testTilesZXY(boundServer, admin, itemId, tileMetadata)
+
+
+@pytest.mark.usefixtures('unbindLargeImage')
+@pytest.mark.plugin('large_image')
+def testTilesConvertRemote(boundServer, admin, fsAssetstore, girderWorker):
+    file = utilities.uploadTestFile('grey10kx5k.tif', admin, fsAssetstore)
+    itemId = str(file['itemId'])
+
+    headers = {
+        'Accept': 'application/json',
+        'Girder-Token': str(Token().createToken(admin)['_id'])
+    }
+    req = requests.post('http://127.0.0.1:%d/api/v1/item/%s/tiles/convert' % (
+        boundServer.boundPort, itemId), headers=headers,
+        data={'localJob': 'false'})
+    assert req.status_code == 200
+    job = req.json()
+    while job['status'] not in (JobStatus.SUCCESS, JobStatus.ERROR, JobStatus.CANCELED):
+        time.sleep(0.1)
+        job = Job().load(job['_id'], force=True)
+    item = Item().findOne({'name': 'grey10kx5k.tiff'}, sort=[('created', SortDir.DESCENDING)])
+    itemId = item['_id']
+    tileMetadata = ImageItem().getMetadata(item)
+    assert tileMetadata['tileWidth'] == 256
+    assert tileMetadata['tileHeight'] == 256
+    assert tileMetadata['sizeX'] == 10000
+    assert tileMetadata['sizeY'] == 5000
+    assert tileMetadata['levels'] == 7
+    assert tileMetadata['magnification'] is None
+    assert tileMetadata['mm_x'] is None
+    assert tileMetadata['mm_y'] is None
+    _testTilesZXY(boundServer, admin, itemId, tileMetadata)
